package ru.dbotthepony.mc.otm.datagen.advancements

import net.minecraft.advancements.Advancement
import net.minecraft.advancements.RequirementsStrategy
import net.minecraft.network.chat.contents.TranslatableContents
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.data.ExistingFileHelper
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.key
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.datagen.lang.MatteryLanguageProvider
import ru.dbotthepony.mc.otm.datagen.modLocation
import ru.dbotthepony.mc.otm.registry.MItemTags
import ru.dbotthepony.mc.otm.registry.MItems
import java.util.function.Consumer

private data class CraftEntry(
	val item: Item,
	val englishName: String,
	val englishSuffix: String? = null,

	val russianName: String? = null,
	val russianSuffix: String? = null,
)

fun addMachineAdvancements(serializer: Consumer<Advancement>, existingFileHelper: ExistingFileHelper, lang: MatteryLanguageProvider) {
	val translation = lang.MultiBuilder("otm.advancements.machine")

	val root = AdvancementBuilder()
		.requirements(RequirementsStrategy.OR)
		.display(
			itemStack = ItemStack(MItems.CHEMICAL_GENERATOR),
			title = translation.add("root", "Tritanium Empowered Machinery") {
				russian("Тританово запитанные механизмы")
			},
			description = translation.add("root.desc", "Do not drop in anything if you want the latter preserved intact") {
				russian("Не роняйте ничего внутрь если хотите чтоб последнее осталось таким, какое оно есть")
			},
			showToast = false,
			announceChat = false,
			background = modLocation("textures/block/decorative/floor_tiles_gray.png")
		)
		.addCriterion("has_machine", criterion(MItemTags.MACHINES))
		.addCriterion("has_tritanium_ingot", criterion(MItemTags.TRITANIUM_INGOTS))
		.addCriterion("has_tritanium_plate_somehow", criterion(MItemTags.TRITANIUM_PLATES))
		.save(serializer, modLocation("machines/root"), existingFileHelper)

	val chem = AdvancementBuilder()
		.parent(root)
		.display(
			itemStack = ItemStack(MItems.CHEMICAL_GENERATOR),
			title = translation.add("chemical_generator", "Burning the Organics") {
				russian("Сжигание органики")
			},
			description = translation.add("chemical_generator.desc", "Craft a Chemical Generator, better to put it outside") {
				russian("Создайте химический генератор. Лучше установить его снаружи")
			},
		)
		.addCriterion("has_machine", criterion(MItems.CHEMICAL_GENERATOR))
		.save(serializer, modLocation("machines/chemical_generator"), existingFileHelper)

	val press = AdvancementBuilder()
		.parent(chem)
		.display(
			itemStack = ItemStack(MItems.PLATE_PRESS),
			title = translation.add("plate_press", "Bending the Material") {
				russian("Раскатка металла")
			},
			description = translation.add("plate_press.desc", "Craft a Plate Press, avoid having limbs inside at all times") {
				russian("Создайте пресс пластин, не суйте свои или чужие конечности внутрь")
			},
		)
		.addCriterion("has_machine", criterion(MItems.PLATE_PRESS))
		.save(serializer, modLocation("machines/plate_press"), existingFileHelper)

	val entries = listOf(
		CraftEntry(MItems.MATTER_SCANNER, "Scanning Things that Matter",
			russianName = "Сканируем вещи которые материальны"),
		CraftEntry(MItems.PATTERN_STORAGE, "Digital Knowledge Library",
			russianName = "Цифровая библиотека знаний"),
		CraftEntry(MItems.MATTER_DECOMPOSER, "Decaying the Atoms", "Keep your limbs outside of the working chamber at all times",
			russianName = "Разлагаем атомы", russianSuffix = "Во всех ситуациях держите свои конечности вне рабочей камеры"),
		CraftEntry(MItems.MATTER_PANEL, "Indexing the Library",
			russianName = "Индексируем библиотеку"),
		CraftEntry(MItems.MATTER_REPLICATOR, "Local Bakery", "Now let's bake some perfect bread",
			russianName = "Местная выпечка", russianSuffix = "А теперь давайте выпечем немного идеального хлеба"),
		CraftEntry(MItems.MATTER_BOTTLER, "Transfusing Pure Matter", "For those who loved to play with water in their childhood",
			russianName = "Переливаем чистую материю", russianSuffix = "Для тех, кто любил играться в воде в детстве"),
		CraftEntry(MItems.MATTER_RECYCLER, "Refine and Redefine", "This is what waste recycling should look like",
			russianName = "Переработка и перегонка", russianSuffix = "Вот он, пик переработки отходов"),
		CraftEntry(MItems.MATTER_CAPACITOR_BANK, "Modular Matter Tank",
			russianName = "Модульный бак материи"),

		CraftEntry(MItems.ENERGY_COUNTER, "Visualize Power Burn",
			russianName = "Визуализация сжигания энергии"),
		CraftEntry(MItems.BATTERY_BANK, "Batteries Not Included", "By all means avoid the urge to hammer incompatible batteries into the power bus.",
			russianName = "Батарейки в комплект не входят", russianSuffix = "Пожалуйста, воздержитесь от вбивания кувалдой несовместимых батарей в энергетическую шину."),
	)

	val built = mutableMapOf<Item, Advancement>()

	for (entry in entries) {
		val path = entry.item.registryName!!.path

		val translated = translation.add("$path.desc", "Craft a %s%s") {
			russian("Создайте %s%s")
		}

		val translatedSuffix = translation.add("$path.suffix", if (entry.englishSuffix != null) ". " + entry.englishSuffix else "") {
			russian(if (entry.russianSuffix != null) ". " + entry.russianSuffix else "")
		}

		built[entry.item] = AdvancementBuilder()
			.parent(press)
			.display(
				itemStack = ItemStack(entry.item),
				title = translation.add(path, entry.englishName) {
					if (entry.russianName != null) {
						russian(entry.russianName)
					}
				},
				description = TranslatableComponent(translated.contents.key, entry.item.description, translatedSuffix),
			)
			.addCriterion("has_machine", criterion(entry.item))
			.save(serializer, modLocation("machines/$path"), existingFileHelper)
	}
}
