package ru.dbotthepony.mc.otm.core.math

import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtAccounter
import net.minecraft.nbt.StringTag
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.util.RandomSource
import net.minecraftforge.common.ForgeConfigSpec
import ru.dbotthepony.mc.otm.config.ObservedConfigValue
import ru.dbotthepony.mc.otm.core.util.readDouble
import ru.dbotthepony.mc.otm.core.util.readVarIntLE
import ru.dbotthepony.mc.otm.core.util.writeDouble
import ru.dbotthepony.mc.otm.core.util.writeVarIntLE
import java.io.InputStream
import java.io.OutputStream
import java.math.BigDecimal
import java.math.BigInteger
import java.math.MathContext
import kotlin.math.absoluteValue

//private fun isZero(value: BigInteger) = value == BigInteger.ZERO
private fun isZero(value: BigInteger) = value.signum() == 0

private fun unsignedInt(value: Byte): Int {
	return value.toInt() and 0xFF
}

private fun usignedLong(value: Byte): Long {
	return value.toLong() and 0xFFL
}

private fun longToBytesBE(value: Long): ByteArray {
	return byteArrayOf(
		(value ushr 56).toByte(),
		(value ushr 48).toByte(),
		(value ushr 40).toByte(),
		(value ushr 32).toByte(),
		(value ushr 24).toByte(),
		(value ushr 16).toByte(),
		(value ushr 8).toByte(),
		(value).toByte(),
	)
}

private fun bytesToLongBE(value: ByteArray, from: Int = 0): Long {
	return (
		(usignedLong(value[from + 7])) or
		(usignedLong(value[from + 6]) shl 8) or
		(usignedLong(value[from + 5]) shl 16) or
		(usignedLong(value[from + 4]) shl 24) or
		(usignedLong(value[from + 3]) shl 32) or
		(usignedLong(value[from + 2]) shl 40) or
		(usignedLong(value[from + 1]) shl 48) or
		(usignedLong(value[from]) shl 56)
	)
}

private fun bytesToLongBE(
	value7: Byte,
	value6: Byte,
	value5: Byte,
	value4: Byte,
	value3: Byte,
	value2: Byte,
	value1: Byte,
	value0: Byte,
): Long {
	return (
		(usignedLong(value0)) or
		(usignedLong(value1) shl 8) or
		(usignedLong(value2) shl 16) or
		(usignedLong(value3) shl 24) or
		(usignedLong(value4) shl 32) or
		(usignedLong(value5) shl 40) or
		(usignedLong(value6) shl 48) or
		(usignedLong(value7) shl 56)
	)
}

/**
 * Constant value which represent edge of meaningful bits.
 *
 * Equals to 0.000000000001
 */
const val EPSILON = 0.000000000001

private val zeroes = Array(16) { "0".repeat(it).intern() }
private val nums = Array(10) { it.toString()[0] }

private fun decimals(input: Double, places: Int): String {
	if (weakEqualDoubles(input, 0.0)) {
		return zeroes.elementAtOrElse(places) { "0".repeat(places) }
	}

	val buffer = CharArray(places)

	@Suppress("NAME_SHADOWING")
	var input = input.absoluteValue

	for (i in 0 until places) {
		input *= 10.0
		buffer[i] = nums[(input % 10.0).toInt()]
	}

	return String(buffer)
}

fun weakEqualDoubles(a: Double, b: Double): Boolean {
	if (a == b)
		return true

	return (a - b).absoluteValue <= EPSILON
}

fun weakCompareDoubles(a: Double, b: Double): Int {
	if (weakEqualDoubles(a, b))
		return 0

	if (a > b)
		return 1

	return -1
}

fun weakLessThan(a: Double, b: Double) = weakCompareDoubles(a, b) < 0
fun weakGreaterThan(a: Double, b: Double) = weakCompareDoubles(a, b) > 0
fun weakLessOrEqual(a: Double, b: Double) = weakCompareDoubles(a, b) <= 0
fun weakGreaterOrEqual(a: Double, b: Double) = weakCompareDoubles(a, b) >= 0

private val PERCENTAGE_CONTEXT = MathContext(6)

private const val MEANINGFUL_BITS_LONG = 1000000000000000000L
private const val MEANINGFUL_BITS_DOUBLE = MEANINGFUL_BITS_LONG.toDouble()
private val MEANINGFUL_BITS_BI = BigInteger.valueOf(MEANINGFUL_BITS_LONG)

private val BI_MINUS_ONE = -BigInteger.ONE

fun Decimal(value: Byte) = Decimal.valueOf(value)
fun Decimal(value: Short) = Decimal.valueOf(value)
fun Decimal(value: Int) = Decimal.valueOf(value)
fun Decimal(value: Long) = Decimal.valueOf(value)

/**
 * Decimal class for having infinitely precise [whole] part, while having fixed precision [decimal] part.
 *
 * In essence, this class is pretty much like [BigDecimal], except decimal part of this number is
 * not guaranteed to be precise (stored as [Double]). The reason behind creation of this class, however,
 * is to allow infinite precision of [whole] part, while leaving [decimal] part to be rounded in inexact operations.
 *
 * This class is value based, hence it is immutable, and can not be utilized as synchronization lock, and should not be compared
 * using identity (`===`) comparison.
 *
 * **In general, this class is lenient about precision**, to avoid gameplay-wise player frustration, rounding is placed in multiple parts of this class,
 * including its constructor (e.g. it is possible to get `Decimal.ZERO == Decimal("0.00000001")`). [EPSILON] is the reference point of precision,
 * any value lesser than [EPSILON] is considered zero (including difference between *the value* and (-)1.0).
 */
@Suppress("unused")
class Decimal @JvmOverloads constructor(whole: BigInteger, decimal: Double = 0.0) : Number(), Comparable<Decimal>, java.io.Serializable {
	constructor(whole: Byte, decimal: Double) : this(BigInteger.valueOf(whole.toLong()), decimal)
	constructor(whole: Short, decimal: Double) : this(BigInteger.valueOf(whole.toLong()), decimal)
	constructor(whole: Int, decimal: Double) : this(BigInteger.valueOf(whole.toLong()), decimal)
	constructor(whole: Long, decimal: Double) : this(BigInteger.valueOf(whole), decimal)
	constructor(value: BigDecimal) : this(value.toBigInteger(), value.remainder(BigDecimal.ONE).toDouble()) // TODO
	constructor(value: Float) : this(BigInteger.valueOf((value - value % 1.0).toLong()), value % 1.0)
	constructor(value: Double) : this(BigInteger.valueOf((value - value % 1.0).toLong()), value % 1.0)
	constructor(value: String) : this(BigDecimal(value))

	/**
	 * Inexact fractional part of this fraction
	 */
	val decimal: Double

	/**
	 * Exact whole part of this fraction
	 */
	val whole: BigInteger

	init {
		@Suppress("name_shadowing")
		var decimal = decimal

		if (decimal < 1.0 && decimal + EPSILON >= 1.0) {
			// разница между дробью и 1 крайне мала
			decimal = 1.0
		} else if (decimal > -1.0 && decimal - EPSILON <= -1.0) {
			// разница между дробью и -1 крайне мала
			decimal = -1.0
		}

		@Suppress("name_shadowing")
		var whole = whole

		// нормализуем
		if (decimal <= -1f || decimal >= 1f) {
			val frac = decimal % 1f
			whole += BigInteger.valueOf((decimal - frac).toLong())
			decimal = frac
		}

		// дробная часть равна или очень близка к нулю
		if (weakEqualDoubles(decimal, 0.0)) {
			this.decimal = 0.0
			this.whole = whole
		// целая часть и дробная часть имеют одинаковые знаки
		} else if (decimal > 0.0 && whole.signum() >= 0 || decimal < 0.0 && whole.signum() <= 0) {
			this.decimal = decimal
			this.whole = whole
		} else {
			// целая часть и дробная часть имеют разные знаки

			if (decimal > 0.0) {
				this.decimal = decimal - 1f
				this.whole = whole + BigInteger.ONE
			//} else if (double < 0.0 && whole.signum() > 0) {
			} else {
				this.decimal = decimal + 1f
				this.whole = whole - BigInteger.ONE
			}

			//throw InternalError("This should never happen, but it happened. Input was: $whole $double")
		}
	}

	val isNaN get() = decimal.isNaN()

	operator fun plus(other: Decimal) = Decimal(whole + other.whole, decimal + other.decimal)
	operator fun minus(other: Decimal) = Decimal(whole - other.whole, decimal - other.decimal)

	operator fun times(other: Decimal): Decimal {
		if (other == ONE)
			return this
		else if (other == MINUS_ONE)
			return -this

		val a = whole
		val c = other.whole

		val bZero = weakEqualDoubles(decimal, 0.0)
		val dZero = weakEqualDoubles(other.decimal, 0.0)

		if (bZero && dZero) {
			if (isZero(a) || isZero(c))
				return ZERO

			return Decimal(a * c)
		} else if (isZero(a) && isZero(c)) {
			if (bZero || dZero)
				return ZERO

			return Decimal(0, decimal * other.decimal)
		} else if (bZero) {
			val d = other.decimal * MEANINGFUL_BITS_DOUBLE
			val dL = d.toLong()
			val adInflated = a * BigInteger.valueOf(dL)
			val ad = adInflated.divideAndRemainder(MEANINGFUL_BITS_BI)

			if (isZero(c)) {
				return Decimal(
					ad[0],
					ad[1].toDouble() / MEANINGFUL_BITS_DOUBLE
				)
			}

			return Decimal(
				a * c + ad[0],
				ad[1].toDouble() / MEANINGFUL_BITS_DOUBLE
			)
		} else if (dZero) {
			val b = decimal * MEANINGFUL_BITS_DOUBLE
			val bL = b.toLong()
			val bcInflated = c * BigInteger.valueOf(bL)
			val bc = bcInflated.divideAndRemainder(MEANINGFUL_BITS_BI)

			if (isZero(a)) {
				return Decimal(
					bc[0],
					bc[1].toDouble() / MEANINGFUL_BITS_DOUBLE
				)
			}

			return Decimal(
				a * c + bc[0],
				bc[1].toDouble() / MEANINGFUL_BITS_DOUBLE
			)
		}

		val b = decimal * MEANINGFUL_BITS_DOUBLE
		val d = other.decimal * MEANINGFUL_BITS_DOUBLE

		val bL = b.toLong()
		val dL = d.toLong()

		val bcInflated = c * BigInteger.valueOf(bL)
		val adInflated = a * BigInteger.valueOf(dL)

		val bc = bcInflated.divideAndRemainder(MEANINGFUL_BITS_BI)
		val ad = adInflated.divideAndRemainder(MEANINGFUL_BITS_BI)

		return Decimal(
			a * c + bc[0] + ad[0],
			decimal * other.decimal + bc[1].toDouble() / MEANINGFUL_BITS_DOUBLE + ad[1].toDouble() / MEANINGFUL_BITS_DOUBLE
		)
	}

	operator fun div(other: Decimal): Decimal {
		if (other == ONE)
			return this
		else if (other == MINUS_ONE)
			return -this

		if (isZero && other.isZero)
			return NaN

		if (!isZero && other.isZero)
			throw ArithmeticException("Divide by zero")

		//if (isZero(whole) && isZero(other.whole))
		//	return ImpreciseFraction(BigInteger.ZERO, decimal / other.decimal)

		val a = toBigDecmial()
		val b = other.toBigDecmial()
		val div = a.divideAndRemainder(b, MathContext(a.precision() + b.precision()))

		val bD = b.toDouble()

		if (bD.isInfinite() || bD == 0.0) {
			return Decimal(div[0].toBigInteger())
		}

		return Decimal(div[0].toBigInteger(), div[1].toDouble() / bD)
	}

	operator fun plus(other: Float): Decimal {
		if (other == 0f) {
			return this
		}

		return plus(Decimal(other))
	}

	operator fun minus(other: Float): Decimal {
		if (other == 0f) {
			return this
		}

		return minus(Decimal(other))
	}

	operator fun times(other: Float): Decimal {
		if (other == 1f) {
			return this
		} else if (other == 0f) {
			return ZERO
		} else if (other == -1f) {
			return -this
		}

		return times(Decimal(other))
	}

	operator fun div(other: Float): Decimal {
		if (other == 0f) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1f) {
			return this
		} else if (other == -1f) {
			return -this
		}

		return div(Decimal(other))
	}

	operator fun plus(other: Double): Decimal {
		if (other == 0.0) {
			return this
		}

		return plus(Decimal(other))
	}

	operator fun minus(other: Double): Decimal {
		if (other == 0.0) {
			return this
		}

		return minus(Decimal(other))
	}

	operator fun times(other: Double): Decimal {
		if (other == 1.0) {
			return this
		} else if (other == 0.0) {
			return ZERO
		} else if (other == -1.0) {
			return -this
		}

		return times(Decimal(other))
	}

	operator fun div(other: Double): Decimal {
		if (other == 0.0) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1.0) {
			return this
		} else if (other == -1.0) {
			return -this
		}

		return div(Decimal(other))
	}

	operator fun plus(other: Int): Decimal {
		if (other == 0) {
			return this
		}

		return plus(valueOf(other))
	}

	operator fun minus(other: Int): Decimal {
		if (other == 0) {
			return this
		}

		return minus(valueOf(other))
	}

	operator fun times(other: Int): Decimal {
		if (other == 1) {
			return this
		} else if (other == 0) {
			return ZERO
		} else if (other == -1) {
			return -this
		}

		return times(valueOf(other))
	}

	operator fun div(other: Int): Decimal {
		if (other == 0) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1) {
			return this
		} else if (other == -1) {
			return -this
		}

		return div(valueOf(other))
	}

	operator fun plus(other: Long): Decimal {
		if (other == 0L) {
			return this
		}

		return plus(valueOf(other))
	}

	operator fun minus(other: Long): Decimal {
		if (other == 0L) {
			return this
		}

		return minus(valueOf(other))
	}

	operator fun times(other: Long): Decimal {
		if (other == 1L) {
			return this
		} else if (other == 0L) {
			return ZERO
		} else if (other == -1L) {
			return -this
		}

		return times(valueOf(other))
	}

	operator fun div(other: Long): Decimal {
		if (other == 0L) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1L) {
			return this
		} else if (other == -1L) {
			return -this
		}

		return div(valueOf(other))
	}

	operator fun plus(other: BigInteger): Decimal {
		if (other == BigInteger.ZERO) {
			return this
		}

		return plus(Decimal(other))
	}

	operator fun minus(other: BigInteger): Decimal {
		if (other == BigInteger.ZERO) {
			return this
		}

		return minus(Decimal(other))
	}

	operator fun times(other: BigInteger): Decimal {
		if (other == BigInteger.ONE) {
			return this
		} else if (other == BigInteger.ZERO) {
			return ZERO
		} else if (other == BI_MINUS_ONE) {
			return -this
		}

		return times(Decimal(other))
	}

	operator fun div(other: BigInteger): Decimal {
		if (other == BigInteger.ZERO) {
			return NaN
		} else if (other == BigInteger.ONE) {
			return this
		} else if (other == BI_MINUS_ONE) {
			return -this
		}

		return div(Decimal(other))
	}

	operator fun unaryMinus(): Decimal {
		return Decimal(-whole, -decimal)
	}

	val absoluteValue: Decimal
		get() {
			if (isNegative) {
				return -this
			} else {
				return this
			}
		}

	override fun equals(other: Any?): Boolean {
		if (isNaN) return false
		return other === this || other is Decimal && other.whole == whole && weakEqualDoubles(decimal, other.decimal)
	}

	override fun hashCode(): Int {
		if (isNaN)
			return Double.NaN.hashCode()

		return 31 * decimal.hashCode() + whole.hashCode()
	}

	fun floor(): Decimal {
		if (decimal == 0.0)
			return this

		return Decimal(whole)
	}

	fun toString(decimals: Int): String {
		if (isNaN)
			return "NaN"

		if (decimals == 0) {
			return whole.toString()
		} else if (decimals > 0) {
			return "$whole.${decimals(decimal, decimals)}"
		}

		return when (signum()) {
			1, -1 -> "$whole.${decimals(decimal, 11)}"
			0 -> "0.0"
			else -> throw IllegalArgumentException("invalid signum")
		}
	}

	override fun toString() = toString(-1)

	fun toBigDecmial(): BigDecimal {
		return BigDecimal(whole) + BigDecimal(decimal)
	}

	private var _signum = Int.MIN_VALUE

	fun signum(): Int {
		if (_signum == Int.MIN_VALUE) {
			var cmp = whole.signum()
			if (cmp == 0) cmp = if (decimal == 0.0) 0 else if (1.0 / decimal > 0.0) 1 else -1
			_signum = cmp
		}

		return _signum
	}

	override fun compareTo(other: Decimal): Int {
		if (isNaN)
			return 1
		else if (other.isNaN)
			return -1

		val a = signum()
		val b = other.signum()

		if (a < b)
			return -1
		else if (a > b)
			return 1

		var cmp = whole.compareTo(other.whole)
		if (cmp == 0) cmp = weakCompareDoubles(decimal, other.decimal)

		return cmp
	}

	fun toByteArray(): ByteArray {
		val whole = whole.toByteArray()
		return byteArrayOf(whole.size.toByte(), *whole, *longToBytesBE(decimal.toBits()))
	}

	fun serializeNBT(): Tag {
		return ByteArrayTag(toByteArray())
	}

	val isZero get() = weakEqualDoubles(decimal, 0.0) && isZero(whole)
	val isPositive get() = this > ZERO
	val isNegative get() = this < ZERO

	fun moreThanZero(): Decimal {
		if (signum() >= 0)
			return this

		return ZERO
	}

	fun lessOrZero(): Decimal {
		if (signum() <= 0)
			return this

		return ZERO
	}

	override fun toInt(): Int {
		if (whole > BI_INT_MAX) {
			return Int.MAX_VALUE
		} else if (whole < BI_INT_MIN) {
			return Int.MIN_VALUE
		}

		return whole.toInt()
	}

	fun write(buff: FriendlyByteBuf) {
		val whole = whole.toByteArray()
		buff.writeByteArray(whole)
		buff.writeDouble(decimal)
	}

	override fun toFloat(): Float {
		return whole.toFloat() + decimal.toFloat()
	}

	override fun toDouble(): Double {
		return whole.toDouble() + decimal
	}

	override fun toByte(): Byte {
		return toInt().toByte()
	}

	override fun toChar(): Char {
		return toInt().toChar()
	}

	override fun toLong(): Long {
		if (whole > BI_LONG_MAX) {
			return Long.MAX_VALUE
		} else if (whole < BI_LONG_MIN) {
			return Long.MIN_VALUE
		}

		return whole.toLong()
	}

	override fun toShort(): Short {
		return toInt().toShort()
	}

	fun percentage(divisor: Decimal, zeroing: Boolean = true): Float {
		if ((isZero || divisor.isZero) && zeroing) return 0f
		if (isNaN || divisor.isNaN || divisor.isZero) return Float.NaN

		if (this >= divisor)
			return 1f

		if (this <= FLOAT_MAX_VALUE && divisor <= FLOAT_MAX_VALUE)
			return (toDouble() / divisor.toDouble()).toFloat()

		return toBigDecmial().divide(divisor.toBigDecmial(), PERCENTAGE_CONTEXT).toFloat()
	}

	operator fun rem(divisor: Decimal): Decimal {
		TODO("Not yet implemented")
	}

	companion object {
		@JvmStatic
		val serialVersionUID: Long = 287354739494574838L

		@JvmField val MINUS_ZERO = Decimal(0, -0.0)
		@JvmField val ZERO = Decimal(BigInteger.ZERO)
		@JvmField val ONE = Decimal(BigInteger.ONE)
		@JvmField val MINUS_ONE = Decimal(-BigInteger.ONE)
		@JvmField val TWO = Decimal(BigInteger.TWO)
		@JvmField val TEN = Decimal(BigInteger.TEN)

		@JvmField val INT_MAX_VALUE = Decimal(BI_INT_MAX)
		@JvmField val INT_MIN_VALUE = Decimal(BI_INT_MIN)
		@JvmField val LONG_MAX_VALUE = Decimal(BI_LONG_MAX)
		@JvmField val LONG_MIN_VALUE = Decimal(BI_LONG_MIN)

		@JvmField val FLOAT_MAX_VALUE = Decimal(BI_FLOAT_MAX)
		@JvmField val FLOAT_MIN_VALUE = Decimal(BI_FLOAT_MIN)
		@JvmField val DOUBLE_MAX_VALUE = Decimal(BI_DOUBLE_MAX)
		@JvmField val DOUBLE_MIN_VALUE = Decimal(BI_DOUBLE_MIN)

		@JvmField val ONE_TENTH = Decimal("0.1")

		private val cache = Array(2048) { Decimal(it - 1024, 0.0) }

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Int): Decimal {
			if (value in -1024 .. 1023) {
				return cache[value + 1024]
			}

			return Decimal(value, 0.0)
		}

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Long): Decimal {
			if (value in -1024L .. 1023L) {
				return cache[value.toInt() + 1024]
			}

			return Decimal(value, 0.0)
		}

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Short) = valueOf(value.toInt())

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Byte) = valueOf(value.toInt())

		@JvmField
		val NaN = Decimal(0, Double.NaN)

		@JvmStatic
		fun fromByteArray(input: ByteArray): Decimal {
			val size = unsignedInt(input[0])

			if (size > 0) {
				val slice = input.copyOfRange(1, 1 + size)
				val bits = bytesToLongBE(input, 1 + size)
				return Decimal(BigInteger(slice), Double.fromBits(bits))
			} else {
				return Decimal(BigInteger.ZERO, Double.fromBits(bytesToLongBE(input, 1)))
			}
		}

		@JvmStatic
		fun deserializeNBT(input: Tag?): Decimal {
			if (input is ByteArrayTag) {
				return fromByteArray(input.asByteArray)
			} else if (input is StringTag) {
				return Decimal(input.asString)
			}

			return ZERO
		}

		@JvmStatic
		fun read(buff: FriendlyByteBuf): Decimal {
			return Decimal(BigInteger(buff.readByteArray()), buff.readDouble())
		}
	}
}

fun FriendlyByteBuf.readDecimal() = Decimal.read(this)
fun FriendlyByteBuf.writeDecimal(value: Decimal) = value.write(this)

fun InputStream.readDecimal(sizeLimit: NbtAccounter = NbtAccounter(512L)): Decimal {
	val size = readVarIntLE(sizeLimit)
	require(size >= 0) { "Negative payload size: $size" }
	sizeLimit.accountBytes(size.toLong() + 8L)
	val bytes = ByteArray(size)
	read(bytes)
	return Decimal(BigInteger(bytes), readDouble())
}

fun OutputStream.writeDecimal(value: Decimal) {
	val bytes = value.whole.toByteArray()
	writeVarIntLE(bytes.size)
	write(bytes)
	writeDouble(value.decimal)
}

fun CompoundTag.getDecimal(key: String) = Decimal.deserializeNBT(this[key])
fun CompoundTag.putDecimal(key: String, value: Decimal) = put(key, value.serializeNBT())

operator fun CompoundTag.set(key: String, value: Decimal) = putDecimal(key, value)

fun Float.toDecimal() = Decimal(this)
fun Double.toDecimal() = Decimal(this)
fun Int.toDecimal() = Decimal(this)
fun Byte.toDecimal() = Decimal(this)
fun Short.toDecimal() = Decimal(this)
fun Long.toDecimal() = Decimal(this)
fun Decimal.toDecimal() = this

class DecimalConfigValue(
	parent: ForgeConfigSpec.ConfigValue<String>,
	val minimum: Decimal? = null,
	val maximum: Decimal? = null,
) : ObservedConfigValue<Decimal>(parent) {
	override fun fromString(value: String): Decimal? {
		try {
			val parsed = Decimal(value)

			if (minimum != null && minimum > parsed) {
				return minimum
			} else if (maximum != null && maximum < parsed) {
				return maximum
			}

			return parsed
		} catch (err: java.lang.NumberFormatException) {
			return null
		}
	}

	override fun toString(value: Decimal): Pair<String, Decimal> {
		if (minimum != null && minimum > value) {
			return minimum.toString() to minimum
		} else if (maximum != null && maximum < value) {
			return maximum.toString() to maximum
		}

		return value.toString() to value
	}
}

private fun ForgeConfigSpec.Builder.commentRange(minimum: Decimal?, maximum: Decimal?) {
	if (minimum != null && maximum != null) {
		comment("Range: $minimum ~ $maximum")
	} else if (minimum != null) {
		comment("Range: >= $minimum")
	} else if (maximum != null) {
		comment("Range: <= $maximum")
	}
}

fun ForgeConfigSpec.Builder.defineDecimal(path: String, defaultValue: Decimal, minimum: Decimal? = null, maximum: Decimal? = null): DecimalConfigValue {
	commentRange(minimum, maximum)
	comment("Default: $defaultValue")
	return DecimalConfigValue(define(path, defaultValue.toString()), minimum, maximum)
}

fun ForgeConfigSpec.Builder.defineDecimal(path: List<String>, defaultValue: Decimal, minimum: Decimal? = null, maximum: Decimal? = null): DecimalConfigValue {
	commentRange(minimum, maximum)
	comment("Default: $defaultValue")
	return DecimalConfigValue(define(path, defaultValue.toString()), minimum, maximum)
}

fun RandomSource.nextDecimal(min: Decimal, max: Decimal, round: Boolean = false): Decimal {
	val value = nextDouble()

	if (round) {
		return Decimal((min + (max - min) * value).whole)
	} else {
		return min + (max - min) * value
	}
}

fun RandomSource.nextVariance(value: Decimal, round: Boolean = false): Decimal {
	return nextDecimal(-value / 2, value / 2, round)
}
