package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceArraySet
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.collect.ProxiedMap
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.util.BigDecimalValueCodec
import ru.dbotthepony.mc.otm.core.util.BinaryStringCodec
import ru.dbotthepony.mc.otm.core.util.BooleanValueCodec
import ru.dbotthepony.mc.otm.core.util.ByteValueCodec
import ru.dbotthepony.mc.otm.core.util.DoubleValueCodec
import ru.dbotthepony.mc.otm.core.util.EnumValueCodec
import ru.dbotthepony.mc.otm.core.util.FloatValueCodec
import ru.dbotthepony.mc.otm.core.util.IStreamCodec
import ru.dbotthepony.mc.otm.core.util.DecimalValueCodec
import ru.dbotthepony.mc.otm.core.util.IntValueCodec
import ru.dbotthepony.mc.otm.core.util.ItemStackValueCodec
import ru.dbotthepony.mc.otm.core.util.LongValueCodec
import ru.dbotthepony.mc.otm.core.util.ShortValueCodec
import ru.dbotthepony.mc.otm.core.util.UUIDValueCodec
import ru.dbotthepony.mc.otm.core.util.VarIntValueCodec
import ru.dbotthepony.mc.otm.core.util.VarLongValueCodec
import ru.dbotthepony.mc.otm.core.util.readVarIntLE
import ru.dbotthepony.mc.otm.core.util.writeVarIntLE
import ru.dbotthepony.mc.otm.secondTime
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.lang.ref.WeakReference
import java.math.BigDecimal
import java.util.*
import java.util.function.Supplier
import kotlin.properties.ReadOnlyProperty
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty
import kotlin.reflect.KProperty0

interface FieldAccess<V> {
	fun read(): V
	fun write(value: V)
}

fun interface FieldGetter<V> {
	fun invoke(field: FieldAccess<V>): V
}

fun interface FieldSetter<V> {
	fun invoke(value: V, access: FieldAccess<V>, setByRemote: Boolean)
}

sealed interface IField<V> : ReadOnlyProperty<Any?, V>, Supplier<V>, () -> V {
	fun observe(): Boolean
	fun markDirty()
	fun markDirty(endpoint: FieldSynchronizer.Endpoint)
	val value: V

	fun write(stream: DataOutputStream, endpoint: FieldSynchronizer.Endpoint)
	fun read(stream: DataInputStream)

	override fun getValue(thisRef: Any?, property: KProperty<*>): V {
		return this.value
	}

	override fun get(): V {
		return value
	}

	override fun invoke(): V {
		return value
	}
}

sealed interface IMutableField<V> : IField<V>, GetterSetter<V> {
	override fun getValue(thisRef: Any?, property: KProperty<*>): V {
		return this.value
	}

	override var value: V

	override fun accept(t: V) {
		value = t
	}

	override fun invoke(): V {
		return this.value
	}
}

data class MapChangeset<out K, out V>(
	val action: MapAction,
	val key: K?,
	val value: V?
)

enum class MapAction {
	CLEAR, ADD, REMOVE
}

/**
 * Universal one-to-many value synchronizer, allowing to synchronize values from server to client
 * anywhere, where input/output streams are supported
 */
@Suppress("unused")
class FieldSynchronizer(private val callback: Runnable, private val alwaysCallCallback: Boolean) {
	constructor() : this(Runnable {}, false)
	constructor(callback: Runnable) : this(callback, false)

	private val fields = ArrayList<AbstractField<*>>(0)
	private val observers = ArrayList<AbstractField<*>>(0)

	private var nextFieldID = 0

	val hasObservers: Boolean get() = observers.isNotEmpty()
	val isEmpty: Boolean get() = fields.isEmpty()
	val isNotEmpty: Boolean get() = fields.isNotEmpty()

	var hasChanges: Boolean = false
		private set(value) {
			if (value != field) {
				field = value

				if (value && !alwaysCallCallback) {
					callback.run()
				}
			}

			if (alwaysCallCallback && value) {
				callback.run()
			}
		}

	fun markClean() {
		hasChanges = false
	}

	fun byte(getter: () -> Byte) = ComputedField(getter, ByteValueCodec)
	fun bool(getter: () -> Boolean) = ComputedField(getter, BooleanValueCodec)
	fun short(getter: () -> Short) = ComputedField(getter, ShortValueCodec)
	fun long(getter: () -> Long) = ComputedField(getter, VarLongValueCodec)
	fun fixedLong(getter: () -> Long) = ComputedField(getter, LongValueCodec)
	fun float(getter: () -> Float) = ComputedField(getter, FloatValueCodec)
	fun double(getter: () -> Double) = ComputedField(getter, DoubleValueCodec)
	fun uuid(getter: () -> UUID) = ComputedField(getter, UUIDValueCodec)
	fun int(getter: () -> Int) = ComputedField(getter, VarIntValueCodec)
	fun fixedInt(getter: () -> Int) = ComputedField(getter, IntValueCodec)
	fun fraction(getter: () -> Decimal) = ComputedField(getter, DecimalValueCodec)
	fun bigDecimal(getter: () -> BigDecimal) = ComputedField(getter, BigDecimalValueCodec)
	fun item(getter: () -> ItemStack) = ComputedField(getter, ItemStackValueCodec)
	fun string(getter: () -> String) = ComputedField(getter, BinaryStringCodec)

	fun byte(getter: KProperty0<Byte>) = ComputedField(getter, ByteValueCodec)
	fun bool(getter: KProperty0<Boolean>) = ComputedField(getter, BooleanValueCodec)
	fun short(getter: KProperty0<Short>) = ComputedField(getter, ShortValueCodec)
	fun long(getter: KProperty0<Long>) = ComputedField(getter, VarLongValueCodec)
	fun fixedLong(getter: KProperty0<Long>) = ComputedField(getter, LongValueCodec)
	fun float(getter: KProperty0<Float>) = ComputedField(getter, FloatValueCodec)
	fun double(getter: KProperty0<Double>) = ComputedField(getter, DoubleValueCodec)
	fun uuid(getter: KProperty0<UUID>) = ComputedField(getter, UUIDValueCodec)
	fun int(getter: KProperty0<Int>) = ComputedField(getter, VarIntValueCodec)
	fun fixedInt(getter: KProperty0<Int>) = ComputedField(getter, IntValueCodec)
	fun fraction(getter: KProperty0<Decimal>) = ComputedField(getter, DecimalValueCodec)
	fun bigDecimal(getter: KProperty0<BigDecimal>) = ComputedField(getter, BigDecimalValueCodec)
	fun item(getter: KProperty0<ItemStack>) = ComputedField(getter, ItemStackValueCodec)
	fun string(getter: KProperty0<String>) = ComputedField(getter, BinaryStringCodec)

	fun byte(getter: Supplier<Byte>) = ComputedField(getter::get, ByteValueCodec)
	fun bool(getter: Supplier<Boolean>) = ComputedField(getter::get, BooleanValueCodec)
	fun short(getter: Supplier<Short>) = ComputedField(getter::get, ShortValueCodec)
	fun long(getter: Supplier<Long>) = ComputedField(getter::get, VarLongValueCodec)
	fun fixedLong(getter: Supplier<Long>) = ComputedField(getter::get, LongValueCodec)
	fun float(getter: Supplier<Float>) = ComputedField(getter::get, FloatValueCodec)
	fun double(getter: Supplier<Double>) = ComputedField(getter::get, DoubleValueCodec)
	fun uuid(getter: Supplier<UUID>) = ComputedField(getter::get, UUIDValueCodec)
	fun int(getter: Supplier<Int>) = ComputedField(getter::get, VarIntValueCodec)
	fun fixedInt(getter: Supplier<Int>) = ComputedField(getter::get, IntValueCodec)
	fun fraction(getter: Supplier<Decimal>) = ComputedField(getter::get, DecimalValueCodec)
	fun bigDecimal(getter: Supplier<BigDecimal>) = ComputedField(getter::get, BigDecimalValueCodec)
	fun item(getter: Supplier<ItemStack>) = ComputedField(getter::get, ItemStackValueCodec)
	fun string(getter: Supplier<String>) = ComputedField(getter::get, BinaryStringCodec)

	fun <T : Enum<T>> enum(type: Class<T>, getter: () -> T) = ComputedField(getter, EnumValueCodec(type))
	inline fun <reified T : Enum<T>> enum(noinline getter: () -> T) = ComputedField(getter, EnumValueCodec(T::class.java))

	fun <T : Enum<T>> enum(type: Class<T>, getter: KProperty0<T>) = ComputedField(getter, EnumValueCodec(type))
	inline fun <reified T : Enum<T>> enum(getter: KProperty0<T>) = ComputedField(getter, EnumValueCodec(T::class.java))

	fun <T : Enum<T>> enum(type: Class<T>, getter: Supplier<T>) = ComputedField(getter::get, EnumValueCodec(type))
	inline fun <reified T : Enum<T>> enum(getter: Supplier<T>) = ComputedField(getter::get, EnumValueCodec(T::class.java))

	@JvmOverloads
	fun byte(
		value: Byte = 0,
		getter: FieldGetter<Byte>? = null,
		setter: FieldSetter<Byte>? = null,

	): Field<Byte> {
		return Field(value, ByteValueCodec, getter, setter)
	}

	@JvmOverloads
	fun bool(
		value: Boolean = false,
		getter: FieldGetter<Boolean>? = null,
		setter: FieldSetter<Boolean>? = null,

	): Field<Boolean> {
		return Field(value, BooleanValueCodec, getter, setter)
	}

	@JvmOverloads
	fun short(
		value: Short = 0,
		getter: FieldGetter<Short>? = null,
		setter: FieldSetter<Short>? = null,

	): Field<Short> {
		return Field(value, ShortValueCodec, getter, setter)
	}

	@JvmOverloads
	fun long(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,

	): Field<Long> {
		return Field(value, VarLongValueCodec, getter, setter)
	}

	@JvmOverloads
	fun fixedLong(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,

	): Field<Long> {
		return Field(value, LongValueCodec, getter, setter)
	}

	@JvmOverloads
	fun float(
		value: Float = 0f,
		getter: FieldGetter<Float>? = null,
		setter: FieldSetter<Float>? = null,

	): Field<Float> {
		return Field(value, FloatValueCodec, getter, setter)
	}

	@JvmOverloads
	fun double(
		value: Double = 0.0,
		getter: FieldGetter<Double>? = null,
		setter: FieldSetter<Double>? = null,

	): Field<Double> {
		return Field(value, DoubleValueCodec, getter, setter)
	}

	@JvmOverloads
	fun uuid(
		value: UUID = UUID(0L, 0L),
		getter: FieldGetter<UUID>? = null,
		setter: FieldSetter<UUID>? = null,

	): Field<UUID> {
		return Field(value, UUIDValueCodec, getter, setter)
	}

	@JvmOverloads
	fun int(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,

	): Field<Int> {
		return Field(value, VarIntValueCodec, getter, setter)
	}

	@JvmOverloads
	fun string(
		value: String = "",
		getter: FieldGetter<String>? = null,
		setter: FieldSetter<String>? = null,

	): Field<String> {
		return Field(value, BinaryStringCodec, getter, setter)
	}

	@JvmOverloads
	fun fixedInt(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,

	): Field<Int> {
		return Field(value, IntValueCodec, getter, setter)
	}

	@JvmOverloads
	fun fraction(
		value: Decimal = Decimal.ZERO,
		getter: FieldGetter<Decimal>? = null,
		setter: FieldSetter<Decimal>? = null,

	): Field<Decimal> {
		return Field(value, DecimalValueCodec, getter, setter)
	}

	@JvmOverloads
	fun bigDecimal(
		value: BigDecimal = BigDecimal.ZERO,
		getter: FieldGetter<BigDecimal>? = null,
		setter: FieldSetter<BigDecimal>? = null,

	): Field<BigDecimal> {
		return Field(value, BigDecimalValueCodec, getter, setter)
	}

	@JvmOverloads
	fun <T : Enum<T>> enum(
		type: Class<T>,
		value: T = type.enumConstants[0],
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,

	): Field<T> {
		return Field(value, EnumValueCodec(type), getter, setter)
	}

	@JvmOverloads
	fun <T : Enum<T>> enum(
		value: T,
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,

	): Field<T> {
		return Field(value, EnumValueCodec(value::class.java), getter, setter)
	}

	@JvmOverloads
	fun item(
		value: ItemStack = ItemStack.EMPTY,
		getter: FieldGetter<ItemStack>? = null,
		setter: FieldSetter<ItemStack>? = null,
		observe: Boolean = true,
	): Field<ItemStack> {
		return Field(value, ItemStackValueCodec, getter, setter, isObserver = observe)
	}

	fun item(
		delegate: KMutableProperty0<ItemStack>,
	): ObservedField<ItemStack> {
		return ObservedField(delegate, ItemStackValueCodec)
	}

	@JvmOverloads
	fun <K, V> map(
		keyCodec: IStreamCodec<K>,
		valueCodec: IStreamCodec<V>,
		callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
		backingMap: MutableMap<K, V> = HashMap(),
		observingBackingMap: MutableMap<K, V>? = null,
	): Map<K, V> {
		return Map(
			keyCodec = keyCodec,
			valueCodec = valueCodec,
			callback = callback,
			backingMap = backingMap,
			observingBackingMap = observingBackingMap,
		)
	}

	private var endpointsMaxCapacity = 1
	private val endpoints = ArrayList<WeakReference<Endpoint>>(1)
	val defaultEndpoint = Endpoint()

	private var nextEndpointsCleanup = secondTime

	private fun notifyEndpoints(dirtyField: AbstractField<*>) {
		hasChanges = true

		forEachEndpoint {
			it.addDirtyField(dirtyField)
		}
	}

	private inline fun forEachEndpoint(execute: (Endpoint) -> Unit) {
		nextEndpointsCleanup = secondTime + 60

		synchronized(endpoints) {
			endpoints.forValidRefs { execute.invoke(it) }

			if (endpoints.size < endpointsMaxCapacity / 2) {
				endpoints.trimToSize()
				endpointsMaxCapacity = endpoints.size
			}
		}
	}

	inner class Endpoint {
		init {
			synchronized(endpoints) {
				endpoints.add(WeakReference(this))
				endpointsMaxCapacity = endpointsMaxCapacity.coerceAtLeast(endpoints.size)

				if (secondTime >= nextEndpointsCleanup) {
					nextEndpointsCleanup = secondTime + 60

					val iterator = endpoints.listIterator()

					for (value in iterator) {
						if (value.get() == null) {
							iterator.remove()
						}
					}

					if (endpoints.size < endpointsMaxCapacity / 2) {
						endpoints.trimToSize()
						endpointsMaxCapacity = endpoints.size
					}
				}
			}
		}

		private val dirtyFields = ReferenceArraySet<AbstractField<*>>(fields.size)

		// use LinkedList because it is ensured memory is freed on LinkedList#clear
		private val mapBacklogs = Reference2ObjectOpenHashMap<Map<*, *>, LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

		var unused: Boolean = false
			private set

		private var mappingVersion = -1

		fun invalidateMappings() {
			mappingVersion = -1
		}

		fun markUnused() {
			require(this === defaultEndpoint) { "This is not a default endpoint" }
			if (unused) return
			unused = true
			mapBacklogs.clear()
			dirtyFields.clear()

			val iterator = endpoints.listIterator()

			for (value in iterator) {
				if (value.get() === this) {
					iterator.remove()
				}
			}
		}

		init {
			markDirty()
		}

		fun markDirty() {
			for (field in fields) {
				field.markDirty(this)
			}
		}

		internal fun addDirtyField(field: AbstractField<*>) {
			if (unused) {
				return
			}

			dirtyFields.add(field)
		}

		internal fun <K, V> getMapBacklog(map: Map<K, V>): LinkedList<Pair<Any?, (DataOutputStream) -> Unit>> {
			if (unused) {
				return LinkedList()
			}

			return mapBacklogs.computeIfAbsent(map, Reference2ObjectFunction {
				LinkedList()
			})
		}

		fun collectNetworkPayload(): FastByteArrayOutputStream? {
			if (unused || dirtyFields.isEmpty()) {
				return null
			}

			val stream = FastByteArrayOutputStream()
			val dataStream = DataOutputStream(stream)

			for (field in dirtyFields) {
				stream.writeVarIntLE(field.id)
				field.write(dataStream, this)
			}

			dirtyFields.clear()
			stream.write(0)

			return stream
		}
	}

	private val boundEndpoints = WeakHashMap<Any, Endpoint>()

	fun computeEndpointFor(obj: Any): Endpoint {
		return boundEndpoints.computeIfAbsent(obj) { Endpoint() }
	}

	fun removeEndpointFor(obj: Any): Endpoint? {
		return boundEndpoints.remove(obj)
	}

	fun endpointFor(obj: Any): Endpoint? {
		return boundEndpoints[obj]
	}

	@Suppress("LeakingThis")
	abstract inner class AbstractField<V> : IField<V> {
		val id: Int = fields.size + 1

		init {
			fields.add(this)
		}
	}

	/**
	 * Networked variable with backing field holding immutable value
	 */
	inner class Field<V>(
		private var field: V,
		private val codec: IStreamCodec<V>,
		private val getter: FieldGetter<V>? = null,
		private val setter: FieldSetter<V>? = null,
		isObserver: Boolean = false,
	) : AbstractField<V>(), IMutableField<V> {
		private var remote: V = codec.copy(field)

		init {
			if (isObserver) {
				observers.add(this)
			}
		}

		private var isDirty = false

		private val access = object : FieldAccess<V> {
			override fun read(): V {
				return field
			}

			override fun write(value: V) {
				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this@Field.field = value
			}
		}

		override fun observe(): Boolean {
			if (!isDirty && !codec.compare(remote, field)) {
				notifyEndpoints(this@Field)
				isDirty = true
			}

			return isDirty
		}

		override var value: V
			get() {
				val getter = this.getter

				if (getter != null) {
					return getter.invoke(access)
				}

				return this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
					return
				}

				if (this.field == value) {
					return
				}

				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this.field = value
			}

		override fun markDirty() {
			notifyEndpoints(this@Field)
			isDirty = true
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			codec.write(stream, field)
			isDirty = false
			remote = codec.copy(field)
		}

		override fun read(stream: DataInputStream) {
			val value = codec.read(stream)
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
				return
			}

			this.field = value
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 */
	inner class ComputedField<V>(
		private val getter: () -> V,
		private val codec: IStreamCodec<V>,
		private val observer: (new: V) -> Unit = {}
	) : AbstractField<V>(), IField<V> {
		private var remote: V? = null
		private var clientValue: V? = null
		private var isDirty = false

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			if (!isDirty && (remote == null || !codec.compare(remote ?: throw ConcurrentModificationException(), value))) {
				notifyEndpoints(this)
				isDirty = true
				remote = codec.copy(value)
			}

			return isDirty
		}

		override fun markDirty() {
			notifyEndpoints(this)
			isDirty = true
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override val value: V
			get() = clientValue ?: getter.invoke()

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			val value = value
			codec.write(stream, value)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			val newValue = codec.read(stream)
			clientValue = newValue
			observer.invoke(newValue)
		}
	}

	/**
	 * Networked variable with backing field holding mutable value, which is constantly observed for changes
	 */
	inner class ObservedField<V> : AbstractField<V>, IMutableField<V> {
		private val codec: IStreamCodec<V>
		private val getter: () -> V
		private val setter: (V) -> Unit
		private var remote: V

		override var value: V
			get() = getter.invoke()
			set(value) { setter.invoke(value) }

		constructor(field: KMutableProperty0<V>, codec: IStreamCodec<V>) : super() {
			this.codec = codec
			getter = field::get
			setter = field::set
			remote = codec.copy(value)
		}

		constructor(getter: () -> V, setter: (V) -> Unit, codec: IStreamCodec<V>) : super() {
			this.codec = codec
			this.getter = getter
			this.setter = setter
			remote = codec.copy(value)
		}

		private var isDirty = false

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			if (!isDirty && !codec.compare(remote, value)) {
				notifyEndpoints(this)
				isDirty = true
				remote = codec.copy(value)
			}

			return isDirty
		}

		override fun markDirty() {
			notifyEndpoints(this)
			isDirty = true
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			val value = value
			codec.write(stream, value)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			this.value = codec.read(stream)
		}
	}

	inner class Map<K, V>(
		private val keyCodec: IStreamCodec<K>,
		private val valueCodec: IStreamCodec<V>,
		private val backingMap: MutableMap<K, V>,
		private val observingBackingMap: MutableMap<K, V>? = null,
		private val callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
	) : AbstractField<MutableMap<K, V>>(), IField<MutableMap<K, V>> {
		private var isDirty = false
		private var sentAllValues = false
		private var isRemote = false

		init {
			if (observingBackingMap != null)
				observers.add(this)
		}

		private fun pushBacklog(key: Any?, value: (DataOutputStream) -> Unit) {
			forEachEndpoint {
				val list = it.getMapBacklog(this)
				val iterator = list.listIterator()

				for (pair in iterator) {
					if (pair.first == key) {
						iterator.remove()
					}
				}

				list.addLast(key to value)
			}
		}

		private fun clearBacklog() {
			forEachEndpoint {
				it.getMapBacklog(this).clear()
			}
		}

		override fun observe(): Boolean {
			if (isRemote) {
				return false
			}

			val observingBackingMap = observingBackingMap

			if (observingBackingMap != null) {
				for ((key, value) in backingMap) {
					val remoteValue = observingBackingMap[key] ?: throw ConcurrentModificationException("Backing map of $this was modified externally, or $value missed a modification")

					if (!valueCodec.compare(value, remoteValue)) {
						val valueCopy = valueCodec.copy(value)

						pushBacklog(key) {
							it.write(MapAction.ADD.ordinal + 1)
							keyCodec.write(it, key)
							valueCodec.write(it, valueCopy)
						}

						observingBackingMap[key] = valueCopy

						if (!isDirty) {
							notifyEndpoints(this)
							isDirty = true
						}
					}
				}
			}

			return isDirty
		}

		override fun markDirty() {
			if (isRemote) {
				return
			}

			if (endpoints.isEmpty()) {
				val observingBackingMap = observingBackingMap ?: return

				for ((key, value) in backingMap)
					observingBackingMap[key] = valueCodec.copy(value)

				return
			}

			isDirty = true
			val backlogs = LinkedList<LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

			forEachEndpoint {
				it.addDirtyField(this)
				val value = it.getMapBacklog(this)
				backlogs.add(value)
				value.clear()
				value.add(null to ClearBacklogEntry)
			}

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				val action = { it: DataOutputStream ->
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				for (backlog in backlogs) {
					backlog.add(key to action)
				}

				observingBackingMap?.put(key, valueCopy)
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			if (isRemote) {
				return
			}

			val backlog = endpoint.getMapBacklog(this)

			backlog.clear()
			backlog.add(null to ClearBacklogEntry)

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				backlog.add(key to {
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				})
			}
		}

		override val value: MutableMap<K, V> = object : ProxiedMap<K, V>(backingMap) {
			override fun onClear() {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				observingBackingMap?.clear()

				forEachEndpoint { endpoint ->
					endpoint.getMapBacklog(this@Map).also {
						it.clear()
						it.add(null to ClearBacklogEntry)
					}
				}

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueAdded(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val valueCopy = valueCodec.copy(value)

				pushBacklog(key) {
					@Suppress("BlockingMethodInNonBlockingContext") // false positive
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				observingBackingMap?.put(key, valueCopy)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueRemoved(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val keyCopy = keyCodec.copy(key)

				pushBacklog(key) {
					@Suppress("BlockingMethodInNonBlockingContext") // false positive
					it.write(MapAction.REMOVE.ordinal + 1)
					keyCodec.write(it, keyCopy)
				}

				observingBackingMap?.remove(key)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			sentAllValues = false
			isDirty = false

			val iterator = endpoint.getMapBacklog(this).listIterator()

			for (entry in iterator) {
				entry.second.invoke(stream)
				iterator.remove()
			}

			stream.write(0)
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				clearBacklog()
				observingBackingMap?.clear()
			}

			isDirty = false

			val changeset = LinkedList<MapChangeset<K, V>>()
			var readAction = stream.read() - 1

			while (readAction != -1) {
				if (readAction >= MapActionList.size) {
					throw IndexOutOfBoundsException("Unknown map action with ID $readAction")
				}

				when (MapActionList[readAction]) {
					MapAction.CLEAR -> {
						backingMap.clear()
						changeset.add(ClearMapChangeset)
					}

					MapAction.ADD -> {
						val key = keyCodec.read(stream)
						val value = valueCodec.read(stream)
						backingMap[key] = value
						changeset.add(MapChangeset(MapAction.ADD, key, value))
					}

					MapAction.REMOVE -> {
						val key = keyCodec.read(stream)
						backingMap.remove(key)
						changeset.add(MapChangeset(MapAction.REMOVE, key, null))
					}
				}

				readAction = stream.read() - 1
			}

			if (changeset.size != 0) {
				callback?.invoke(changeset)
			}
		}
	}

	/**
	 * marks all fields dirty, invalidates mappings for each endpoint
	 */
	fun invalidate() {
		for (field in fields) {
			field.markDirty()
		}

		forEachEndpoint {
			it.invalidateMappings()
		}
	}

	/**
	 * Observe changes of all fields with backing computation lambda
	 */
	fun observe(): Boolean {
		var changes = false

		if (observers.isNotEmpty()) {
			for (field in observers) {
				changes = field.observe() || changes
			}
		}

		return changes
	}

	/**
	 * [defaultEndpoint]#collectNetworkPayload
	 */
	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		check(!defaultEndpoint.unused) { "Default endpoint is not used" }
		observe()
		val values = defaultEndpoint.collectNetworkPayload()
		markClean()
		return values
	}

	fun read(stream: InputStream): Int {
		var fieldId = stream.readVarIntLE()
		var i = 0
		val dataStream = DataInputStream(stream)

		while (fieldId != 0) {
			val field = fields.getOrNull(fieldId - 1) ?: throw IllegalArgumentException("Unknown field ID ${fieldId - 1}")
			field.read(dataStream)
			fieldId = stream.readVarIntLE()
			i++
		}

		return i
	}

	companion object {
		private val ClearBacklogEntry = { stream: DataOutputStream -> stream.write(MapAction.CLEAR.ordinal + 1) }
		private val MapActionList = MapAction.values()
		private val ClearMapChangeset = MapChangeset(MapAction.CLEAR, null, null)
	}
}
